! Program to process .txt chain files produced by CosmoMC (cosmologist.info/cosmomc)
! Calculates statistics, thins, produces 1D and 2D marginalized plots and scatter plots
! and writes out useful information.
! Option to adjust priors and map parameters - write code below

!Output files are (depending on options)

! * file_root.margestats - the mean, limits and stddev from the 1D marginalized distributions
! * file_root.likestats  - the best fit, and limits from the N-D confidence region
! * file_root.sm         - SuperMongo .sm file to produce 1D marginalized plots
! * file_root_2D.m       - MatLab .m file to produce 2D marginalized plots
! * file_root_3D.m       - MatLab .m file to produce 2D sample plots colored by a third parameter
! * file_root_thin.txt   - thinned merged versions of input files
! * file_root_single.txt - File of weight-1 single samples generated by probabilistic importance sampling
! * file_root.covmat     - covariance matrix for parameters
! * file_root.corr       - covariance matrix for parameters normalized to have unit variance
! * file_root.PCA        - human-readable file giving details of PCA, constrained parameters, etc

! You can easily edit the .m and .sm files produced to produce custom layouts of plots, etc.
! Data for plots are exported to the plot_data folder

!This version June 2002

module MCSamples
 use settings
 implicit none
        real, dimension(:,:), target, allocatable :: coldata  !(col_index, row_index)
	
        integer, parameter :: max_rows = 300000
        integer, parameter :: max_cols = 30
        integer nrows, ncols
        real numsamp
        integer ND_cont1, ND_cont2
	integer covmat_dimension
        integer colix(max_cols), num_vars  !Parameters with non-blank lables
        real mean(max_cols), sddev(max_cols)
        real, dimension(:,:), allocatable :: corrmatrix
	character(LEN=120) rootname
        character(LEN=120) lables(max_cols)
contains

  subroutine AdjustPriors
   !Can adjust the multiplicity of each sample in coldata(1, rownum) for new priors
   !Be careful as this code is parameterisation dependent
   integer i
   real chisq
   real H0, Age, e1, ns
   real s8, omm

   stop 'You need to write the AdjustPriors subroutine in GetDist.F90 first!'

   write (*,*) 'Adjusting priors'
   do i=0, nrows-1

   !e.g. put in HST prior..
   
     chisq = (H0-72)**2/(8**2)  
    
     coldata(1,i) = coldata(1,i)*exp(-chisq/2) 
     coldata(2,i) = coldata(2,i) + chisq/2
      
   end do
   

  end subroutine AdjustPriors

  subroutine MapParameters(invars)
    real invars(1:ncols)
    real H0, h2, omm, s8, As,h, fv, omdmh2, tau, age, zre, ombh2
  
      stop 'Need to write MapParameters routine first'  

        As = invars(11+2)
        H0 = invars(3+2)
	omdmh2 = invars(2+2)
        ombh2 = invars(1+2)
        fv = invars(6+2)
        age = invars(15+2)
	omm = invars(16+2)
        zre = invars(4+2)
        h=H0/100
	s8 = invars(17+2)
        tau = invars(18+2)

  end subroutine MapParameters

  subroutine CoolChain(cool)
    real, intent(in) :: cool
    integer i
    real maxL, newL
    
    write (*,*) 'Cooling chains by ', cool
    MaxL = minval(coldata(2,0:nrows-1))
    do i=0, nrows-1
     newL = coldata(2,i)*cool
     coldata(1,i) = coldata(1,i)*exp(-(newL - coldata(2,i)) - MaxL*(1-cool) )
     coldata(2,i) = newL
   end do

  end subroutine CoolChain

  subroutine SortColData(bycol)
 !Sort coldata in order of likelihood
     Type(Real_pointer), dimension(:), allocatable :: rowptrs
     integer, intent(in) :: bycol
     real, dimension(:,:), pointer :: tmp 
     integer i

     allocate(tmp(ncols,0:nrows-1))
  
     allocate(rowptrs(0:nrows-1))
     do i = 0, nrows -1
      rowptrs(i)%p => coldata(1:,i)
     end do

    call QuickSortArr(rowptrs(0), 1, nrows, bycol)

!Make new table using sorted pointers	
    do i = 0, nrows-1
	tmp(1:ncols,i) = rowptrs(i)%p(1:ncols)
    end do

    deallocate(rowptrs)
    coldata(1:ncols,0:nrows-1) = tmp(1:ncols,0:nrows-1)
    deallocate(tmp)

  end  subroutine SortColData


  subroutine MakeSingleSamples

  !Make file of weight-1 samples by choosing samples with probability given by their weight

    use Random
    integer i
    character(LEN=20) :: fmt
    real*8 ranmar
    external ranmar
	
    real maxmult         
    call initRandom()
    fmt = trim(numcat('(',ncols))//'E15.5)'
    open(unit=50,file=trim(rootname)//'_single.txt',form='formatted',status='replace')
    maxmult = maxval(coldata(1,0:nrows-1))
    do i= 0, nrows -1
     if (ranmar() <= coldata(1,i)/maxmult) write (50,fmt) 1.0, coldata(2:ncols,i) 	       
    end do         
    close(50)

  end subroutine MakeSingleSamples

  subroutine ThinData(fac)
  !Make thinned chain .txt file
 
    integer, intent(in) :: fac
    integer i, tot, nout, mult
    character(LEN=20) :: fmt
             
    fmt = trim(numcat('(',ncols))//'E15.5)'
    open(unit=50,file=trim(rootname)//'_thin.txt',form='formatted',status='replace')
              
    tot= 0
    nout=0
    i = 0

    write (*,*) 'thinning ', nrows, 'lines of output'

    mult = coldata(1,0)
    do while (i< nrows)
        
    if (abs(nint(coldata(1,i)) - coldata(1,i)) > 1e-4) stop 'non-integer weights in ThinData'

    if (mult + tot < fac) then
       tot = tot + mult
       i=i+1
       if (i< nrows) mult = nint(coldata(1,i))
    else
       write (50,fmt) 1.0, coldata(2:ncols,i)
       
       nout= nout+1
       if (mult == fac - tot) then  
          i=i+1
	  if (i< nrows) mult = nint(coldata(1,i))
       else
         mult = mult - (fac -tot)
       end if
       tot = 0
    end if

    end do
       
    write (*,*) 'output ',nout, 'samples'

    close(50)

  end subroutine ThinData

  
	  subroutine GetCovMatrix
	    integer i, j
	    real mean(max_cols)
	    real onecol(0:max_rows)
            real scale
            real, dimension(:,:), allocatable :: covmatrix
         
             allocate(corrmatrix(ncols-2,ncols-2))
	
	     corrmatrix = 0

             do i=1, ncols-2
               mean(i) = sum(coldata(1,0:nrows-1)*coldata(i+2,0:nrows-1))/numsamp
             end do

	     do i = 1, ncols-2
              do j = i, ncols-2
	         corrmatrix(i,j) = sum(coldata(1,0:nrows-1)*((coldata(2+i,0:nrows-1)-mean(i))* &
                                            (coldata(2+j,0:nrows-1)-mean(j))))/numsamp
                 corrmatrix(j,i) = corrmatrix(i,j)
	      end do
	     end do	

             if (covmat_dimension /= 0) then
              write (*,*) 'Writing covariance matrix for ',covmat_dimension,' parameters'
              allocate(covmatrix(covmat_dimension,covmat_dimension))
	      covmatrix=corrmatrix(1:covmat_dimension,1:covmat_dimension)
	      call WriteSqMatrix(trim(rootname) //'.covmat',covmatrix, covmat_dimension)
              deallocate(covmatrix)
             end if

              do i=1, ncols-2
               if (corrmatrix(i,i) > 0) then
               scale = sqrt(corrmatrix(i,i))
               corrmatrix(i,:) = corrmatrix(i,:)/scale
               corrmatrix(:,i) = corrmatrix(:,i)/scale
               end if
             end do
             call WriteSqMatrix(trim(rootname) //'.corr',corrmatrix, ncols -2)
             
          end subroutine GetCovMatrix


          function MostCorrelated2D(i1,i2, direc)
            integer, intent(in) :: i1,i2, direc
            integer MostCorrelated2D
            !Find which parameter is most correllated with the degeneracy in ix1, ix2
            integer pars(2)
            real mat2D(2,2), evals(2), u(2,2)
            real corrs(2,ncols-2)
          
            if (direc /= 0 .and. direc /= -1) stop 'Invalid 3D color parameter'

            pars(1)=i1
            pars(2)=i2
            mat2D = corrmatrix(pars,pars)
            u = mat2D

            call Diagonalize(u, evals, 2)
            corrs = matmul(transpose(u), corrmatrix(pars,:))
            corrs(:,pars) = 0
          
            MostCorrelated2D = MaxIndex(abs(corrs(2+direc,colix(1:num_vars)-2)), num_vars)
            MostCorrelated2D = colix(MostCorrelated2D) -2
         
          end function MostCorrelated2D

           function ConfidVal(ix,limfrac,upper)
             integer, intent(IN) :: ix
             real, intent(IN) :: limfrac
             logical, intent(IN) :: upper
             real ConfidVal
             real try, lasttry, try_t, try_b

          !find upper and lower bounds
             try_b = minval(coldata(ix,0:nrows-1))
             try_t = maxval(coldata(ix,0:nrows-1))
            
             lasttry = -1
             if (upper) then
             do 
               try = sum(coldata(1,0:nrows-1),mask = coldata(ix,0:nrows-1) > (try_b + try_t)/2) 
               if (try > numsamp*limfrac) then
                  try_b = (try_b+try_t)/2
               else
                  try_t = (try_b+try_t)/2
               end if
               if (try == lasttry) exit
               lasttry = try
             end do
             else
               do 
               try = sum(coldata(1,0:nrows-1),mask = coldata(ix,0:nrows-1) < (try_b + try_t)/2) 
               if (try > numsamp*limfrac) then
                  try_t = (try_b+try_t)/2
               else
                  try_b = (try_b+try_t)/2
               end if
               if (try == lasttry) exit
               lasttry = try
                end do
             end if

             ConfidVal = try_t
            end function ConfidVal

          subroutine PCA(pars,n,param_map,normparam_num)
            !Perform principle component analysis
            !In other words, get eigenvectors and eigenvalues for normalized variables
            !with optional (log) mapping
	    integer, intent(in) :: n, pars(n), normparam_num
            character(LEN=*), intent(IN) ::  param_map
            integer i, j, normparam, locarr(1:1)
            character(LEN =30) fmt

	    real PCmean(n), sd(n), newmean(n), newsd(n)
	    real evals(n)
            real corrmatrix(n,n), u(n,n)
            real, dimension(:,:), allocatable :: PCdata
            character(LEN=100) PClabs(n), div, expo
            character(LEN=1000) tmp
            logical doexp

	     write (*,*) 'Doing PCA for ',n,' parameters'
             call CreateTxtFile(trim(rootname) //'.PCA', 40)
             write (40,*) 'PCA for parameters: '

             if (normparam_num /=0) then
                normparam = IndexOf(normparam_num,pars,n)
                if (normparam ==0) stop 'Invalid PCA normalization parameter'
             else
                normparam = 0
             end if

             allocate(PCdata(n,0:nrows-1))
             
             PCdata(:,:) = coldata(2+pars,0:nrows-1)
             fmt = trim(numcat('',n))//'f8.4)'

             doexp =.false.
	     do i = 1, n
                if (param_map(i:i)=='L') then 
                   doexp = .true.
                   PCdata(i,:) = log(PCdata(i,:))
                   PClabs(i) = 'ln(' //trim(lables(pars(i)+2))//')'
                elseif (param_map(i:i)=='M') then 
                   doexp = .true.
                   PCdata(i,:) = log(-PCdata(i,:))
                   PClabs(i) = 'ln(-' //trim(lables(pars(i)+2))//')'
                else 
                   PClabs(i) = trim(lables(pars(i)+2))
                end if
                write (40,*) pars(i),':',trim(PClabs(i))
                PCmean(i) = sum(coldata(1,0:nrows-1)*PCdata(i,:))/numsamp
                PCdata(i,:) = PCdata(i,:) - PCmean(i)
                sd(i) = sqrt(sum(coldata(1,0:nrows-1)*PCdata(i,:)**2)/numsamp)
                if (sd(i)/=0) PCdata(i,:) = PCdata(i,:)/sd(i)
                corrmatrix(i,i) = 1
             end do

             write (40,*) ''
             write (40,*) 'Correlation matrix for reduced parameters'
             do i = 1, n
              do j = i+1, n
	          corrmatrix(i,j) = sum(coldata(1,0:nrows-1)*PCdata(i,:)*PCdata(j,:))/numsamp
                  corrmatrix(j,i) = corrmatrix(i,j)
	      end do
                write (40,'(1I4,": ",'//trim(fmt)) pars(i), corrmatrix(i,:)
	     end do	
	        
             u = corrmatrix
             call Diagonalize(u, evals, n)

             write (40,*) ''
             write (40,*) 'e-values of correlation matrix'
             do i = 1, n
                write (40,'("PC",1I2,": ",'//trim(fmt)) i, evals(i)
             end do

             write (40,*) ''
             write (40,*) 'e-vectors'
             
             do i = 1, n
                write (40,'(1I3,": ",'//trim(fmt)) pars(i), u(i,:)
             end do

          
              if (normparam /= 0) then
                !Set so parameter normparam has exponent 1
                do i=1, n
                   u(:,i) = u(:,i)/u(normparam,i)*sd(normparam)
                end do
              else
                !Normalize so main component has exponent 1
                do i=1, n
                   locarr(1:1) = maxloc(abs(u(:,i)))
                   u(:,i) = u(:,i)/u(locarr(1),i)*sd(locarr(1))
                end do
              end if
                     
             do i = 0, nrows -1
               PCdata(:,i) = matmul(transpose(u), PCdata(:,i))
               if (doexp) PCdata(:,i) = exp(PCdata(:,i))
             end do

             write (40,*) ''
             write (40,*) 'Principle components'
            
             do i = 1, n
                write (40,*) trim(numcat('PC',i))//' (e-value: '//trim(RealToStr(evals(i)))//')'
                do j=1,n
                   if (param_map(j:j)=='L' .or. param_map(j:j)=='M' ) then
                     expo = RealToStr(1/sd(j)*u(j,i) )
                    if (param_map(j:j)=='M') then
                      div = RealToStr( -exp(PCmean(j)))
                    else
                      div = RealToStr( exp(PCmean(j)))
                    end if
                    write(40,*) '['//trim(RealToStr(u(j,i)))//']   ('//trim(lables(pars(j)+2)) &
                          //'/'//trim(div)//')^{'//trim(expo)//'}'
                   else
                      expo = RealToStr(sd(j)/u(j,i))
                      if (doexp) then
                       write(40,*) '['//trim(RealToStr(u(j,i)))//']    exp(('// &
                        trim(lables(pars(j)+2))//'-'//trim(RealToStr(PCmean(j)))// ')/'// trim(expo)//')'
                      else
                        write(40,*) '['//trim(RealToStr(u(j,i)))//']   ('// &
                       trim(lables(pars(j)+2))//'-'//trim(RealToStr(PCmean(j)))// ')/'// trim(expo)
                      end if
                   end if

                end do
                newmean(i) = sum(coldata(1,0:nrows-1)*PCdata(i,:))/numsamp
                newsd(i) = sqrt(sum(coldata(1,0:nrows-1)*(PCdata(i,:)-newmean(i))**2)/numsamp) 
                write (40,*) '          = '//trim(RealToStr(newmean(i)))// ' +- '//trim(RealToStr(newsd(i)))
                write (40,'(" ND limits: ",4f9.3)') minval(PCdata(i,0:ND_cont1)),maxval(PCdata(i,0:ND_cont1)), &
                                             minval(PCdata(i,0:ND_cont2)),maxval(PCdata(i,0:ND_cont2))
                write (40,*) ''
                
             end do

!Find out how correlated these components are with other parameters
              write (40,*) 'Correlations of principle components'
   
              write (40,trim(numcat('("    ",',n))//'I8)') (I, I=1, n)

              do i=1, n
               PCdata(i,:) = (PCdata(i,:) - newmean(i)) /newsd(i)
              end do

               do j=1,n
                  write (40,'("PC",1I2)', advance = 'no') j
                  do i=1,n
                   write (40,'(1f8.3)', advance ='no') sum(coldata(1,0:nrows-1)*PCdata(i,:)* &
                       PCdata(j,:))/numsamp
                  end do
                  write (40,*) ''
               end do

               do j=1, num_vars
                   write (40,'(1I4)', advance = 'no') colix(j)-2
                   do i=1,n
                    write (40,'(1f8.3)', advance ='no') sum(coldata(1,0:nrows-1)*PCdata(i,:)* &
                      (coldata(colix(j),0:nrows-1)-mean(j))/sddev(j))/numsamp
                   end do
                   write (40,*) '  ('//trim(lables(colix(j)))//')'
               end do


             close(40)
             deallocate(PCdata)

          end subroutine PCA
 
end module MCSamples


program GetDist

        use IniFile

        use MCSamples
        implicit none
      
        character(LEN=80) InputFile, numstr
        character(LEN=300) BestFit, fname
        character(LEN=10000) InLine
        character(LEN=80) ComparePlots(20)
        integer Num_ComparePlots
        real  bincounts(-1000:1000),binlikes(-1000:1000),binmaxlikes(-1000:1000)
	real bins2D(-1000:1000,-1000:1000),bin2Dlikes(-1000:1000,-1000:1000),bin2Dmax(-1000:1000,-1000:1000)
        real, dimension(:,:), allocatable :: TheBins
          !allow non-integer, e.g. after importance sampling
        real onecol(0:max_rows), invars(max_cols)
        integer ix, ix1,i,ix2,ix3, nbins, ignorerows, wx,wy
        logical BW

        logical bad, adjust_priors

        integer method, numbins, ix_min(max_cols),ix_max(max_cols)
        real center(max_cols),width(max_cols), val, amax, amin
        character(LEN=120) filename, infile, in_root,out_root, sm_file, fmt, plotfile
        character(LEN=120) matlab_col
        integer plot_row, plot_col, chain_num, chain_ix, plot_num
       
        integer x,y,j,j2, i2,num_2D_plots, num_cust2D_plots
        integer num_3D_plots
        character(LEN=80) plot_3D(20)
        integer plot_x, plot_y, thin_factor, outliers
        real mean_mult, max_mult
        logical has_limits(max_cols)
        real limmin(max_cols),limmax(max_cols), maxbin
        integer plot_2D_param, j2min
        real try, lasttry
        real try_b, try_t,try_sum, try_last
        real cont_lines(max_cols,2,2), dist, distweight,norm, limfrac 
        real meanlike, binweight, maxlike, maxmult
        real contours(2), contour_levels(2)
	integer bestfit_ix
        integer num_contours, chain_exclude
        logical smoothing, plot_meanlikes, plot_NDcontours
	logical map_params, shade_meanlikes
	logical make_single_samples
	integer cust2DPlots(max_cols**2)
	real counts
        real cool
        integer PCA_num, PCA_normparam
        integer PCA_params(max_cols)
        integer plotparams_num, plotparams(max_cols)
        character(LEN=max_cols) PCA_func

	InputFile = GetParam(1)
        if (InputFile == '') stop 'No parameter input file'

        call Ini_Open(InputFile, 1, bad, .false.)
        if (bad) stop 'Error opening parameter file'
        Ini_fail_on_not_found = .true.

        rootname = Ini_Read_String('file_root')
	in_root = rootname
        chain_num = Ini_Read_Int('chain_num')

        ncols = Ini_Read_Int('columnnum',0)

        thin_factor = Ini_Read_Int('thin_factor',0)
	
        allocate(coldata(ncols,0:max_rows))

        nbins = Ini_Read_Int('num_bins')

        ignorerows = Ini_Read_Int('ignore_rows',0)

        adjust_priors = Ini_Read_Logical('adjust_priors',.false.)
        
        lables(1) = 'mult'
        lables(2) = 'likelihood'
        do ix=3, ncols
           write (numstr,*) ix-2
           lables(ix) = Ini_Read_String('lab'//trim(adjustl(numstr)))
        end do

        Ini_fail_on_not_found = .false.
	make_single_samples = Ini_Read_logical('make_single_samples',.false.)
        cool = Ini_Read_Real('cool',1.)

        Num_ComparePlots = Ini_Read_Int('compare_num',0)
        do ix = 1, Num_ComparePlots
           ComparePlots(ix) = Ini_Read_String(numcat('compare',ix))
        end do

        has_limits = .false.
        do ix=3, ncols
           write (numstr,*) ix-2
           InLine = Ini_Read_String('limits'//trim(adjustl(numstr)))
           if (InLine /= '') then
              has_limits(ix) = .true.
              read (InLine,*) limmin(ix),limmax(ix)
           end if
        end do

        plot_2D_param = Ini_Read_Int('plot_2D_param',0)
        if (plot_2D_param /= 0) then
	  plot_2D_param = plot_2D_param + 2
	  num_cust2D_plots = 0
	else
         !Use custom array of specific plots
	 num_cust2D_plots = Ini_Read_Int('plot_2D_num',0)
	 do ix = 1, num_cust2D_plots
	  	InLine = Ini_Read_String(numcat('plot',ix))
	        read (InLine, *) ix1,ix2
	        cust2DPLots(ix) = ix1+2 + (ix2+2)*100 
	 end do
	
	end if

	chain_exclude = Ini_read_Int('exclude_chain',0)
	map_params = Ini_Read_Logical('map_params',.false.)
	if (map_params)  &
          write (*,*) 'WARNING: Mapping params - .covmat file is new params.'

	shade_meanlikes = Ini_Read_Logical('shade_meanlikes',.false.)	

        matlab_col = Ini_Read_String('matlab_colscheme')
        smoothing = Ini_Read_Logical('smoothing',.true.)

	out_root = Ini_Read_String('out_root')
	if (out_root /= '') then
	 rootname = out_root
	 write (*,*) 'producing files with root '//trim(out_root)
	end if
	sm_file = trim(rootname) // '.sm'


        contours(1) = Ini_Read_Real('contour1',0.)
        contours(2) = Ini_Read_Real('contour2',0.)
        num_contours = count(contours/= 0)
        if (contours(1)==0) contours(1) = contours(2)


	covmat_dimension = Ini_Read_Int('cov_matrix_dimension',0)

	plot_meanlikes = Ini_Read_Logical('plot_meanlikes',.false.)
	plot_NDcontours = Ini_Read_Logical('plot_NDcontours',.false.)

        PCA_num = Ini_Read_Int('PCA_num',0)
        if (PCA_num /= 0) then
           if (PCA_num <2) stop 'Can only do PCA for 2 or more parameters'
           InLine = Ini_Read_String('PCA_params')
           PCA_func =Ini_Read_String('PCA_func')
            !Characters representing functional mapping
           if (PCA_func == '') PCA_func(1:PCA_num) = 'N' !no mapping
           read (InLine,*) PCA_params(1:PCA_num)
           PCA_NormParam = Ini_Read_Int('PCA_normparam',0)
        end if

        plotparams_num = Ini_Read_Int('plotparams_num',0)
        if (plotparams_num /= 0) then
          InLine = Ini_Read_String('plotparams')
          read(InLine,*) plotparams(1:plotparams_num)
        end if

        num_3D_plots = Ini_Read_Int('num_3D_plots',0)
        do ix =1, num_3D_plots
           plot_3D(ix) = Ini_Read_String(numcat('3D_plot',ix))
           make_single_samples = .true.
        end do

        BW = Ini_Read_Logical('B&W',.false.)

        call Ini_Close


!Read in the chains 	
        nrows = 0

        do chain_ix = 1, max(1,chain_num)

  	 if (chain_ix == chain_exclude .and. chain_num /= 0) cycle
         if (chain_num == 0) then
            infile = trim(in_root) // '.txt'
         else
            write (numstr,*) chain_ix
            infile = trim(in_root) //'_'//trim(adjustl(numstr))// '.txt'
         end if

         write (*,*) 'reading ' // trim(infile)
         open(unit=50,file=infile,form='formatted',status='old')

          do ix = 1, ignorerows
             read (50,'(a)',end=1) InLine
          end do

          do
           read (50,'(a)',end=1) InLine
           read(InLine,*,end=1, err=2) invars(1:ncols) 
           if (map_params) call MapParameters(invars)
           coldata(1:ncols, nrows) = invars(1:ncols)
           
           nrows = nrows + 1
           if (nrows > max_rows) stop 'need to increase max_rows'

          end do    
 2       write (*,*) 'error reading line ', nrows + ignorerows
 	 stop
 1       close(50)

          end do
          if (nrows == 0) stop 'No un-ignored rows!'

          if (cool /= 1) call CoolChain(cool)
!Adjust weights if requested
          if (adjust_priors) call AdjustPriors
	
	
          mean_mult = sum(coldata(1,0:nrows-1))/nrows
          write (*,*) 'mean multiplicity = ',mean_mult

          max_mult = (mean_mult*nrows)/min(nrows/2,500)
          outliers = count(coldata(1,0:nrows-1) > max_mult)
          write (*,*) 'outlier fraction ', real(outliers)/nrows

!Output thinned data if requested
!Must do this with unsorted output	
    	  if (thin_factor /= 0) then 
             call ThinData(thin_factor)
          end if

!Produce file of weight-1 samples if requested
	  if (make_single_samples) call MakeSingleSamples

!Sort data in order of likelihood of points

          call SortColData(2)
      
          numsamp = sum(coldata(1,0:nrows-1))

!Get ND confidence region (index into sorted coldata)
          counts = 0
	  ND_cont1=-1; ND_cont2=-1
	  do j=0, nrows -1
	    counts = counts + coldata(1,j)
	    if (counts > numsamp*contours(1) .and. ND_cont1==-1) then
	     ND_cont1 = j
             if (ND_cont2 /= -1) exit
	    end if
	    if (counts > numsamp*contours(2) .and. ND_cont2==-1) then
	     ND_cont2 = j
	    end if
	  end do          

!Only use variables whose lable's are not empty (and in list of plotparams if plotparams_num /= 0)
          num_vars = 0
          
          do ix = 3,ncols
             if (lables(ix) /= '') then
                if (plotparams_num == 0 .or. any(plotparams(1:plotparams_num)==ix-2)) then
                 num_vars = num_vars + 1
                 colix(num_vars) = ix
                 mean(num_vars) = sum(coldata(1,0:nrows-1)*coldata(ix,0:nrows-1))/numsamp
                 sddev(num_vars)  = sqrt(sum(coldata(1,0:nrows-1)*(coldata(ix,0:nrows-1) &
                           -mean(num_vars))**2)/numsamp)
                end if
             end if
          end do

          maxmult = maxval(coldata(1,0:nrows-1))

          write (*,*) 'using ',nrows,' rows, processing ',num_vars,' parameters'
          write (*,*) 'effective number of samples (assuming indep): ', nint(numsamp/maxmult)

!Get covariance matrix and correlation matrix
          call GetCovMatrix

          if (PCA_num>0) call PCA(PCA_params,PCA_num,PCA_func, PCA_NormParam)

!Find best fit, and mean likelihood     
	  bestfit_ix = 0 !Since we have sorted the lines
	  maxlike = coldata(2,bestfit_ix)
	  write (*,*) 'Best fit -Ln(like) = ', maxlike

          meanlike = sum(coldata(2,0:nrows-1)*coldata(1,0:nrows-1)) / numsamp 
	  write (*,*) 'mean(-Ln(like)) = ', meanlike
	  
	  meanlike = -log(sum(exp(-(coldata(2,0:nrows-1) -coldata(2,bestfit_ix)))*coldata(1,0:nrows-1)) & 
                  / numsamp) + coldata(2,bestfit_ix)
	  write (*,*) '-Ln(mean like) = ', meanlike
	  	

!Output .sm file for 1D plots

          plot_col =  nint(sqrt(num_vars/1.4))
          plot_row = (num_vars +plot_col-1)/plot_col 
   
          !So we don't forget to run sm, delete current .ps file
	  call DeleteFile(trim(rootname)//'.ps')
    
          open(unit=50,file=sm_file,form='formatted',status='replace')
          write (50,*) 'DEFINE TeX_strings 1'
          write (50,*) 'device postfile '//trim(rootname)//'.ps'
          write (50,*) 'DEFINE default_font rm'
          if (num_vars > 6) then
           write (50,*) 'expand 0.8'
           write (50, *)'define x_gutter 0.6'
           write (50,*) 'define y_gutter 0.8'
          else
           write (50,*) 'expand 1.1'
          endif

          write (50,*) 'location 2500 31000 3500 31000'
          if (BW) then
            write (50,*) 'lweight 4'
          else
             write (50,*) 'lweight 3'
          end if

          cont_lines = 0

!Do 1D bins
          do j = 1,num_vars
             ix = colix(j)
             bincounts = 0
             binlikes = 0
	     binmaxlikes = 0
             
             do ix1 = 1, num_contours

              limfrac = (1-contours(ix1))
	      if (.not. has_limits(ix)) limfrac = limfrac/2 !two tail
              cont_lines(j,2,ix1) = ConfidVal(ix,limfrac,.true.)
              cont_lines(j,1,ix1) = ConfidVal(ix,limfrac,.false.)

             end do !contour lines

             if (has_limits(ix)) then
              amin = limmin(ix)
              amax = limmax(ix)
             else
              amax = ConfidVal(ix,0.0001,.true.)
              amin = ConfidVal(ix,0.0001,.false.)
             end if
             width(j) = (amax-amin)/(nbins+1)
             if (amin <= 0) then 
               center(j) = amax 
             else
               center(j) = amin 

             end if

             ix_min(j) = nint((amin - center(j))/width(j))
             ix_max(j) = nint((amax - center(j))/width(j))

	     if (smoothing .and. .not. has_limits(ix) .and. amin > 0) ix_min(j) = ix_min(j)-1
	     if (smoothing .and. .not. has_limits(ix)) ix_max(j) = ix_max(j)+1

             do i = 0, nrows-1
                ix2=nint((coldata(ix,i)-center(j))/width(j))
                if (smoothing .and. ix_min(j) /= ix_max(j)) then
                 do wx = max(ix_min(j),ix2-2),min(ix_max(j),ix2+2)
                   dist = coldata(colix(j),i) - (center(j) + wx*width(j))
	           distweight = exp(- dist**2/width(j)**2/2)
                   binweight = coldata(1,i)*distweight
                   bincounts(wx) = bincounts(wx)  + binweight
                   binlikes(wx) = binlikes(wx) + binweight*exp(meanlike-coldata(2,i))
	           binmaxlikes(wx) = max(binmaxlikes(wx), exp(maxlike-coldata(2,i))*distweight)
                 end do
               else
                 bincounts(ix2) = bincounts(ix2) + coldata(1,i)
	         binlikes(ix2) = binlikes(ix2) + coldata(1,i)*exp(meanlike-coldata(2,i))
	         binmaxlikes(wx) = max(binmaxlikes(wx), exp(maxlike-coldata(2,i)))
               end if
             end do
             
	     do i=ix_min(j),ix_max(j)
	      if (bincounts(i) >0) binlikes(i) = binlikes(i)/bincounts(i)
	     end do

             if (has_limits(ix) .and. smoothing .and. ix_min(j) /= ix_max(j)) then
                !account for underweighting near edges
                bincounts(ix_min(j))   = bincounts(ix_min(j))*2
                bincounts(ix_max(j))   = bincounts(ix_max(j))*2
                bincounts(ix_max(j)-1) = bincounts(ix_max(j)-1)/0.84
                bincounts(ix_min(j)+1) = bincounts(ix_min(j)+1)/0.84
             end if
      

          filename = 'plot_data/'//trim(rootname)//trim(numcat('_p',colix(j)-2))
          open(unit=49,file=trim(filename)//'.dat',form='formatted',status='replace')
          maxbin = maxval(bincounts(ix_min(j):ix_max(j)))
          do i = ix_min(j), ix_max(j)
             write (49,'(3E15.5)') center(j) + i*width(j), bincounts(i)/maxbin, binmaxlikes(i)
          end do
          if (ix_min(j) == ix_max(j)) write (49,'(1E15.5," 0")') center(j) + ix_min*width(j)
          close(49)

	  if (plot_meanlikes) then
	 
	   open(unit=49,file=trim(filename)//'.likes',form='formatted',status='replace')	   
	   maxbin = maxval(binlikes(ix_min(j):ix_max(j)))
	   do i = ix_min(j), ix_max(j)
             write (49,'(2E15.5)') center(j) + i*width(j), binlikes(i)/maxbin
           end do
	   close(49)
	  end if


          plot_x = mod(j+plot_col-1,plot_col)+1
          plot_y = plot_row - (j+plot_col-1)/plot_col + 1 

          if (lables(ix) /= '') then
           write (50,'("window",4I4)') plot_col,plot_row, plot_x, plot_y
          
   
           write (50,*) 'data "' // trim(filename)//'.dat"'
           write (50,*) 'read { x 1 y 2 mx 3}'
           if (ix_min(j) == ix_max(j)) then
              write (50,'("limits ",2E15.5," y")') center(j) - abs(center(j))/2-0.1, &
                          center(j) + abs(center(j))/2+0.1
           else
             if (has_limits(ix)) then
              write (50,'("limits ",2E15.5," 0 1.1")') amin, amax
             else
              write (50,*) 'limits x 0 1.1'
             end if
           end if
           write (50,*) 'box 1 3 3 3'
           
           write (50,*) 'connect x y'
!	   write (50,*) 'connect x mx '
	   !uncomment to plot maximum likelihood values for each bin

	   if (plot_meanlikes) then
	    write (50,*) 'data "' // trim(filename)//'.likes"'
            write (50,*) 'read { x 1 y 2 }'
	    write (50,*) 'ltype 1'
	    write (50,*) 'connect x y'
	   end if
         
           do ix1 = 1, Num_ComparePlots
             fname =  trim('plot_data/'//trim(ComparePlots(ix1))// &
                               trim(numcat('_p',colix(j)-2)))//'.dat'
    
             if (FileExists(fname)) then
             write (50,*) 'data "' // trim(fname) //'"'
             write (50,*) 'read { x 1 y 2 mx 3}'
              write (50,*) 'ltype 0'
             if (BW) then
               write (50,*) 'lweight 2'
               if (ix1>1) write (50,*) 'ltype 2'
             else
               write (50,*) 'ctype ',ix1+2
             end if
          
             write (50,*) 'connect x y'
             if (plot_meanlikes) then
                write (50,*) 'data "' // trim('plot_data/'//trim(ComparePlots(ix1))// &
                     trim(numcat('_p',colix(j)-2)))//'.likes"'
                write (50,*) 'read { x 1 y 2}'
                write (50,*) 'ltype 1'
        	write (50,*) 'connect x y'
             end if
             end if
           end do
           if (Num_ComparePlots/=0) write (50,*) 'ctype black'
           if (Num_ComparePlots/=0 .and. BW) write (50,*) 'lweight 4'


  !         write (50,*) 'ltype 1'
  !         do ix1 = 1, num_contours
  !         do i2 =1,2
  !         write (numstr,*) cont_lines(j,i2,ix1)
  !         write (50,*) 'set x = {'//trim(numstr)//' '//trim(numstr)//'}'
  !         write (50,*) 'set y = { 0 1.1}'
  !!         write (50,*) 'connect x y'
  !uncomment this if you want dotted lines on 1D plots for condifence regions
  !Note 1D confidence regions are computed from full samples using different definition to
  !2D confidence regions

  !         end do
  !         end do

           write (50,*) 'ltype 0'
           write (50,*) 'xlabel '// trim(lables(ix))
           write (50,*)
           write (50,*) '#Next...'
           
          end if
          end do

   
          write (50,*) 'quit'
          close(50)
                 

!do 2D bins and produce matlab file
      
          if (plot_2D_param == 0 .and. num_cust2D_plots==0) then
             !In this case output the most correlated variable combinations
          
             write (*,*) 'doing 2D plots for most correlated variables'
             try_t = 1e5
            
             num_cust2D_plots = 12
              do j = 1, num_cust2D_plots
               try_b = -1e5
               do ix1 =1 ,num_vars
                do ix2 = ix1+1,num_vars 
                   if (abs(corrmatrix(colix(ix1)-2,colix(ix2)-2)) < try_t .and. &
                        abs(corrmatrix(colix(ix1)-2,colix(ix2)-2)) > try_b) then
                      try_b = abs(corrmatrix(colix(ix1)-2,colix(ix2)-2))
                      x = ix1; y = ix2;
                    end if
                end do
               end do
               if (try_b == -1e5) then
                  num_cust2D_plots = j-1
                  exit
               end if
               try_t = try_b

             cust2Dplots(j) = colix(x) + colix(y)*100
             end do

          end if


	  if (num_cust2D_plots == 0) then
           num_2D_plots = 0
        
           do j= 1, num_vars
             if (ix_min(j) /= ix_max(j)) then
             do j2 = j+1, num_vars
              if (ix_min(j2) /= ix_max(j2)) then
                if (plot_2D_param==0 .or. plot_2D_param == colix(j) .or. plot_2D_param == colix(j2)) &
                      num_2D_plots = num_2D_plots + 1
              end if
            end do
            end if
           end do
	  else
	   num_2D_plots = num_cust2D_plots
	  end if
 
         write (*,*) 'Producing ',num_2D_plots,' 2D plots'
     
         if (num_2D_plots /= 0) then

          filename = trim(rootname)//'_2D.m'
          open(unit=50,file=filename,form='formatted',status='replace')
     
          if (num_2D_plots <7) then
            write(50,*) 'lab_fontsize = 10; axes_fontsize = 10;'
   	   else
            write(50,*) 'lab_fontsize = 9; axes_fontsize = 9;'
          end if
          plot_col = nint(sqrt(num_2D_plots/1.4))
          plot_row = (num_2D_plots +plot_col-1)/plot_col 
          plot_num = 0
	
          do j= 1, num_vars
           if (ix_min(j) /= ix_max(j)) then
            
            if (plot_2D_param/=0 .or. num_cust2D_plots /= 0) then
               if (colix(j) == plot_2D_param) cycle
               j2min = 1
            else
               j2min = j+1
            end if

           do j2 = j2min, num_vars
              if (ix_min(j2) /= ix_max(j2)) then

              if (plot_2D_param/=0 .and. colix(j2) /= plot_2D_param) cycle
	      if (num_cust2D_plots /= 0 .and.  &
                count(cust2Dplots(1:num_cust2D_plots) == colix(j)*100 + colix(j2))==0) cycle
	   
              plot_num = plot_num + 1

              
              bins2D = 0
	      bin2Dlikes = 0
	      bin2Dmax = 1e30        

              do i = 0, nrows-1
                ix1=nint((coldata(colix(j),i)-center(j))/width(j))
                ix2=nint((coldata(colix(j2),i)-center(j2))/width(j2))
       
                if (smoothing) then
                do wx = max(ix_min(j),ix1-2),min(ix_max(j),ix1+2)
                   dist = (coldata(colix(j),i) - (center(j) + wx*width(j)))**2/width(j)**2
                 do wy = max(ix_min(j2),ix2-2),min(ix_max(j2),ix2+2)
		   binweight = coldata(1,i)*exp( - (&
                      dist +(coldata(colix(j2),i) - (center(j2) + wy*width(j2)))**2/width(j2)**2 )/2)
                   bins2D(wx,wy) = bins2D(wx,wy) + binweight
	           bin2Dlikes(wx,wy) = bin2Dlikes(wx,wy) + binweight*exp(meanlike - coldata(2,i))
                 end do
                end do
                else
                  bins2D(ix1,ix2) = bins2D(ix1,ix2) + coldata(1,i) 
	          bin2Dlikes(ix,ix2) = bin2Dlikes(ix,ix2) + coldata(1,i)*exp(meanlike-coldata(2,i))
                end if
	          bin2Dmax(ix1,ix2) = min(bin2Dmax(ix1,ix2),coldata(2,i))
              end do

             
              do ix1=ix_min(j),ix_max(j)
               do ix2 =ix_min(j2),ix_max(j2) 
	        if (bins2D(ix1,ix2) >0) bin2Dlikes(ix1,ix2) = bin2Dlikes(ix1,ix2)/bins2D(ix1,ix2)
		 if (plot_NDcontours) then
                 !Map maximum likelihood in each bin into a significance value from the full N-D distribution
                  if (bin2DMax(ix1,ix2) == 1e30) then
	            bin2DMax(ix1,ix2) = 0
	          else
	            bin2DMax(ix1,ix2) = sum(coldata(1,0:nrows-1), mask = coldata(2,0:nrows-1) > bin2DMax(ix1,ix2))/numsamp
	          end if
	         end if
	       end do 
	      end do
	 

              if (has_limits(colix(j)) .or. has_limits(colix(j2))) then

               !Fix up underweighting near edges. Note this makes edge pixels noisier.
               do ix1 = ix_min(j), ix_max(j)
                 do ix2 = ix_min(j2),ix_max(j2)
                    if (ix1 ==ix_min(j)) bins2D(ix1,ix2) = bins2D(ix1,ix2)*2
                    if (ix1 ==ix_max(j)) bins2D(ix1,ix2) = bins2D(ix1,ix2)*2
                    if (ix2 ==ix_min(j2)) bins2D(ix1,ix2) = bins2D(ix1,ix2)*2
                    if (ix2 ==ix_max(j2)) bins2D(ix1,ix2) = bins2D(ix1,ix2)*2

                    if (ix1 ==ix_min(j)+1) bins2D(ix1,ix2) = bins2D(ix1,ix2)/0.84
                    if (ix1 ==ix_max(j)-1) bins2D(ix1,ix2) = bins2D(ix1,ix2)/0.84
                    if (ix2 ==ix_min(j2)+1) bins2D(ix1,ix2) = bins2D(ix1,ix2)/0.84
                    if (ix2 ==ix_max(j2)-1) bins2D(ix1,ix2) = bins2D(ix1,ix2)/0.84
                 end do
               end do

             end if

! Get contour containing contours(:) of the probability

             allocate(TheBins(ix_max(j)-ix_min(j)+1,ix_max(j2)-ix_min(j2)+1))
             TheBins = bins2D(ix_min(j):ix_max(j),ix_min(j2):ix_max(j2))
                
             norm = sum(TheBins)
             
             do ix1 = 1, num_contours
             try_t = maxval(TheBins)
             try_b = 0
             try_last = -1
             do
               try_sum = sum(TheBins,mask = TheBins < (try_b + try_t)/2) 
               if (try_sum > (1-contours(ix1))*norm) then
                  try_t = (try_b+try_t)/2
               else
                  try_b = (try_b+try_t)/2
               end if
               if (try_sum == try_last) exit
               try_last = try_sum
         
             end do
             contour_levels(ix1) = (try_b+try_t)/2

             end do
             deallocate(TheBins)



              plotfile = numcat(trim(numcat(trim(rootname)//'_2D_',colix(j)-2))//'_',colix(j2)-2)
              filename = 'plot_data/'//trim(plotfile)
              open(unit=49,file=filename,form='formatted',status='replace')
              do ix1 = ix_min(j), ix_max(j)
                  write (49,trim(numcat('(',ix_max(j2)-ix_min(j2)+1))//'E15.5)') bins2D(ix1,ix_min(j2):ix_max(j2))
              end do
              close(49)

	      if (shade_meanlikes) then
                  open(unit=49,file=trim(filename)//'_likes',form='formatted',status='replace')
	          maxbin = maxval(bin2Dlikes(ix_min(j):ix_max(j),ix_min(j2):ix_max(j2)))
                  do ix1 = ix_min(j), ix_max(j)
                    write (49,trim(numcat('(',ix_max(j2)-ix_min(j2)+1))//'E15.5)') &
	                       bin2Dlikes(ix1,ix_min(j2):ix_max(j2))/maxbin
                  end do
                 close(49)
	      end if

	      if (plot_NDcontours) then
                  open(unit=49,file=trim(filename)//'_confid',form='formatted',status='replace')
                  do ix1 = ix_min(j), ix_max(j)
                    write (49,trim(numcat('(',ix_max(j2)-ix_min(j2)+1))//'E15.5)') &
	                       bin2Dmax(ix1,ix_min(j2):ix_max(j2))

                  end do
                 close(49)
	      end if

              if (plot_2D_param==0 .or. plot_2D_param == colix(j) .or. plot_2D_param == colix(j2)) then
 
              write (50,*) 'load '//trim(filename)//';'
	      if (shade_meanlikes) write (50,*) 'load '//trim(filename)//'_likes;'
	      if (plot_NDcontours) write (50,*) 'load '//trim(filename)//'_confid;'
              fmt = trim(numcat('("x1 = [",',ix_max(j2)-ix_min(j2)+1))//'E15.5,"];")'
              write (50,fmt) (/(I, I=ix_min(j2), ix_max(j2), 1)/)*width(j2)+center(j2)
              fmt = trim(numcat('("x2 = [",',ix_max(j)-ix_min(j)+1))//'E15.5,"];")'
              write (50,fmt) (/(I, I=ix_min(j), ix_max(j), 1)/)*width(j)+center(j)

              write (50,'("subplot(",1I5,",",1I5,",",1I5,");")') plot_row,plot_col,plot_num
      	      if (shade_meanlikes) then
	        write (50,*) 'contourf(x1,x2,'//trim(plotfile)//'_likes,64);'
    	      else     
                write (50,*) 'contourf(x1,x2,'//trim(plotfile)//',64);'
              end if

              write (50,*) 'shading flat; hold on;'
              if (num_contours /= 0) then
               write(numstr,*) contour_levels(1:num_contours)
               if (num_contours==1) numstr = trim(numstr)//' '//trim(numstr)
               write (50,*) '[C h] = contour(x1,x2,'//trim(plotfile)//',['//trim(numstr)//'],''k'');'

	       if (plot_NDcontours) then
	        write(numstr,*) 1-contours(1:num_contours)  
	        if (num_contours==1) numstr = trim(numstr)//' '//trim(numstr)
	        write (50,*) '[C h] = contour(x1,x2,'//trim(plotfile)//'_confid,['//trim(numstr)//'],''r'');'
	       end if
              end if
             
              write (50,*) 'hold off; set(gca,''Layer'',''top'',''FontSize'',axes_fontsize);'
                        
              fmt = ''',''FontSize'',lab_fontsize);'
              write (50,*) 'xlabel('''//trim(lables(colix(j2)))//trim(fmt)
              write (50,*) 'ylabel(''',trim(lables(colix(j)))//trim(fmt)
              
              end if
          
              end if
           end do
           end if
          end do
	 
          if (matlab_col/='') write (50,*) trim(matlab_col)
	  if (num_2D_plots < 5) then
            write (50,*) 'set(gcf, ''PaperPosition'',[ 0 0 6 6]);';
	  else
	    write (50,*) 'set(gcf, ''PaperPosition'',[ 0 0 8 10]);';
	  end if
          write (50,*) 'print -dpsc2 '//trim(rootname)//'_2D.ps;'
          close(50)

          end if

!Do 3D plots (i.e. 2D scatter plots with coloured points)

          if (num_3D_plots /=0) then
              write (*,*) 'producing ',num_3D_plots, '2D colored scatter plots'
              filename = trim(rootname)//'_3D.m'
              open(unit=50,file=filename,form='formatted',status='replace')
              write (50,*) 'colormap(''jet'');'
              if (num_3D_plots ==1 ) then
                write(50,*) 'lab_fontsize = 12; axes_fontsize = 12;'
              else
                write(50,*) 'lab_fontsize = 9; axes_fontsize = 9;'
              end if
              if (mod(num_3D_plots,2)==0 .and. num_3D_plots < 11) then
               plot_col = num_3D_plots/2
              else
               plot_col =  nint(sqrt(1.*num_3D_plots))
              end if
              plot_row = (num_3D_plots +plot_col-1)/plot_col 

              write (50,*) 'set(gcf, ''PaperPosition'',[ 0 0 '//  &
                           trim(RealToStr((plot_col*9.)/plot_row)) //' 8]);'

              write(50,*) 'load '//trim(rootname)//'_single.txt;'
              write (50,*) 'pts = '//trim(rootname)//'_single;'
             
              do j=1, num_3D_plots
                 read(plot_3D(j),*) ix1, ix2, ix3  !x, y, color
                 if (ix3<1) ix3 = MostCorrelated2D(ix1,ix2,ix3)
                 
                 write (50,'("subplot(",1I5,",",1I5,",",1I5,");")') plot_row,plot_col,j   
                 write (50,*) '%Do params ',ix1,ix2,ix3
                 write (50,*) 'scatter(pts(:,',2+ix1,'),pts(:,',2+ix2,'),3,pts(:,',2+ix3,'));'
                 fmt = ''',''FontSize'',lab_fontsize);'
                 write (50,*) 'xlabel('''//trim(lables(ix1+2))//trim(fmt)
                 write (50,*) 'ylabel('''//trim(lables(ix2+2))//trim(fmt)
                 write (50,*) 'set(gca,''FontSize'',axes_fontsize); pa = gca;'
                 write (50,*) 'hbar = colorbar(''horiz'');axes(hbar);'
                 write (50,*) 'xlabel('''//trim(lables(ix3+2))//trim(fmt)
                 write (50,*) 'set(gca,''FontSize'',axes_fontsize);'
                 write (50,*) ' p = get(pa,''Position'');'
                 if (num_3D_plots > 2) write (50,*) 'set(pa,''Position'',[p(1) (p(2)+p(4)/10) p(3) p(4)]);'
              end do
              write (50,*) 'print -dpsc2 '//trim(rootname)//'_3D.ps;'
              close(50)
          end if
    

!write out stats

          open(unit=50,file=trim(rootname)//'.margestats',form='formatted',status='replace')
	  write(50,*) 'param  mean        sddev       lower1      upper1      lower2      upper2'
          do j=1, num_vars
             write(50,'(1I5,6E12.3)') colix(j)-2, mean(j), sddev(j), &
                  cont_lines(j,1:2,1),cont_lines(j,1:2,2)
          end do
          close(50)



!Limits from global likelihood
	  open(unit=50,file=trim(rootname)//'.likestats',form='formatted',status='replace')
	  write(50,*) 'param  bestfit     lower1      upper1      lower2      upper2'
	 	  
          do j=1, num_vars
	 
             write(50,'(1I5,5E12.3)') colix(j)-2, coldata(colix(j),bestfit_ix),&
              minval(coldata(colix(j),0:ND_cont1)), &
 	      maxval(coldata(colix(j),0:ND_cont1)), &
 	      minval(coldata(colix(j),0:ND_cont2)), &
 	      maxval(coldata(colix(j),0:ND_cont2))
          end do
          close(50)



end program GetDist

