!Module storing observed datasets, their points and window functions
!and routines for computing the likelihood
!Cls passed to these routines are in in MicroK^2, no 2pi or l(l+1)

!Windows functions read in from files are assumed to be W_l/l. The band powers are obtained from
!W_l by \sum_l (2l+1) W_l C_l/2pi where W_l is normalized so \sum_l (2l+1)/l(l+1) W_l = 1.

!This code is very indebted to Sarah's cosmarg
!Analytic marginalization over calibration and beam from astro-ph/0112114

!x factors are described in astro-ph/9808264. Essentially they are a first correction
!for the bandpower errors being non-Gaussian. If no x-factors we assume they are gaussian.
!Using x-factors we transform to a variable Z = log (bandpower + x), and assume Z has Gaussian errors.
!When x-factors are used the inverse covariance matrix is assumed to be that for Z.
!See the RADPACK page for data http://bubba.ucdavis.edu/~knox/radpack.html

!Hyperparameters follow astro-ph/0203259. We preserve normalization relative to the no hyperparameter 
!case so that the evidence ratios can be compared. Hyperparameters do not work properly unless
!no x-factors.

!This version July 2002


module cmbdata
use settings
use cmbtypes
implicit none

 
!if CMBdataset%has_xfactors then obs, var and N_inv are for the offset-lognormal variables Z

  Type CMBdatapoint
    integer win_min, win_max
     !Ranges in which window is non-zero
    real, pointer, dimension(:) :: window
     !Normalized window function in l
    real obs, err_minus, err_plus, sigma, var
     !Observed value of l(l+1) C_l/2pi bandpowers in MicroK^2, with errors
    real beam_err !fractional beam error (file is value in MicroK^2)
  end Type CMBdatapoint

  Type CMBdataset
    logical :: use_set
    real :: calib_uncertainty
    logical :: beam_uncertain, has_corr_errors, has_xfactors
    logical :: use_hyperparameter
    integer :: num_points, file_points
    character(LEN=20) :: name
    Type(CMBdatapoint), pointer, dimension(:) :: points
    real, pointer, dimension(:,:) :: N_inv
    real, pointer, dimension(:) :: xfactors
    real :: hyper_norm !Log(Gamma(n/2 +1) 2^(n/2+1))
  end Type CMBdataset

  integer :: num_datasets = 0
  logical :: CMB_hyperparameters = .false.
  Type(CMBdataset) datasets(10)
 
  integer, parameter :: halfsteps = 5 !do 2*halfsteps+1 steps in numerical marginalization
  real margeweights(-halfsteps:halfsteps)
  real :: margenorm = 0 
  private halfsteps, margeweights, margenorm
contains

 subroutine CombineSets(S1,S2, calib)
   !I use this for fudging VSA and CBI data with correlated calibrations
   integer i,j
   real, intent(in) :: calib
   character(LEN=*), intent(in) :: S1, S2
   Type(CMBdataset) S(2)
   Type(CMBdataset) aset
   integer ix,x

   i=-1; j=-1
   do ix=1, num_datasets
      if (datasets(ix)%name == S1) i = ix
      if (datasets(ix)%name == S2) j = ix
   end do

   if (i==-1 .or. j==-1) return

   if (Feedback> 0) write (*,*) 'Combining datasets '//trim(S1)//' and '//trim(S2)

   S(1)= datasets(i)
   S(2)= datasets(j)

   do ix = 1,2
      if (.not. S(ix)%has_xfactors) then
         allocate(S(ix)%xfactors(S(ix)%num_points))
         S(ix)%xfactors = 5000
         S(ix)%has_xfactors = .true.
         S(ix)%points(:)%var = S(ix)%points(:)%var/(S(ix)%points(:)%obs +S(ix)%xfactors)**2
         S(ix)%points(:)%obs = log(S(ix)%points(:)%obs +S(ix)%xfactors)
         if (S(ix)%has_corr_errors) stop 'expected no N_inv for merged non-xfactor dataset'
      end if

      if (.not. S(ix)%has_corr_errors) then
         S(ix)%has_corr_errors = .true.
         allocate(S(ix)%N_inv(S(ix)%num_points,S(ix)%num_points ))
         S(ix)%N_inv = 0
         do x=1, S(ix)%num_points
            S(ix)%N_inv(x,x) = 1/S(ix)%points(x)%var
         end do
      end if
   end do
   
  
 !  aset = S(1)
   aset%use_set = .true.
   aset%beam_uncertain = .false.
   if (S(1)%beam_uncertain .or. S(2)%beam_uncertain) stop 'Combining not supported with beam uncertainties'
   aset%has_corr_errors = .true.
   aset%has_xfactors = .true.
   aset%use_hyperparameter = .false.

   aset%num_points = sum(S(:)%num_points)
   allocate(aset%points(aset%num_points))
   aset%points(1:S(1)%num_points) = S(1)%points
   aset%points(S(1)%num_points+1 :S(1)%num_points+S(2)%num_points) = S(2)%points
   allocate(aset%xfactors(aset%num_points))
   aset%xfactors(1:S(1)%num_points) = S(1)%xfactors
   aset%xfactors(S(1)%num_points+1 :S(1)%num_points+S(2)%num_points) = S(2)%xfactors
   allocate(aset%N_inv(aset%num_points,aset%num_points))
   aset%N_inv = 0
   aset%N_inv(1:S(1)%num_points,1:S(1)%num_points) = S(1)%N_inv
   aset%N_inv(S(1)%num_points+1 :S(1)%num_points+S(2)%num_points, &
              S(1)%num_points+1 :S(1)%num_points+S(2)%num_points) = S(2)%N_inv

   aset%name = trim(S(1)%name) // trim(S(2)%name)
   aset%calib_uncertainty = calib

   datasets(i) = aset
   datasets(j)%use_set = .false.
 end subroutine CombineSets

 subroutine ReadWindow(AP, aname, are_bare)
   Type(CMBdatapoint) AP
   character(LEN=*), intent(IN) :: aname
   logical, intent(IN) :: are_bare
   integer l
   real w,ll, IW

   if (Feedback > 1) write (*,*) 'reading window: '//trim(aname)
   allocate(AP%window(lmax))
   AP%window = 0
   call OpenTxtFile(aname, 50)

     do
     read(50,*, end=1) ll, w
     l=nint(ll)
     if (abs(l-ll) > 1e-4) stop 'non-integer l in window file' 
     if (l>=2 .and. l<=lmax) then
        AP%window(l) = w
        if (.not. are_bare) AP%window(l) = AP%window(l)*l
     else
        if (l>lmax .and. w /= 0) then
           write (*,*) 'Error: Window function outside l_max: ',trim(aname)
           stop
        end if
     end if
     end do
1  close(50)

   do l=2, lmax
      if (AP%window(l)/=0) then
         AP%win_min = l
         exit
      end if
   end do

   do l=lmax,2,-1
      if (AP%window(l)/=0) then
         AP%win_max = l
         exit
      end if
   end do
   IW = 0
   do l = AP%win_min, AP%win_max
      IW = IW + ((2*l+1)*AP%window(l))/(l*(l+1))
   end do
   
   AP%window(AP%win_min:AP%win_max) = AP%window(AP%win_min:AP%win_max)/IW

 end subroutine ReadWindow


 subroutine ReadDataset(aname)
   character(LEN=*), intent(IN) :: aname
   character(LEN=120) :: InLine, window_dir, Ninv_file, xfact_file
   logical bad, windows_are_bare
   Type (CMBdataset) :: aset
   integer i
   real, pointer, dimension(:,:) :: tmp_mat

   num_datasets = num_datasets + 1
   if (num_datasets > 10) stop 'too many datasets'

   call Ini_Open(aname, 1, bad, .false.)
   if (bad) then
     write (*,*)  'Error opening dataset file '//trim(aname)
     stop
   end if
   Ini_fail_on_not_found = .true.
      
   aset%name = Ini_Read_String('name')
   aset%use_set =.true.
    
   if (Feedback > 0) write (*,*) 'reading: '//trim(aset%name)

   aset%num_points = Ini_Read_Int('num_points')

   aset%hyper_norm = LogGamma(aset%num_points/2.+1) + (aset%num_points/2.0+1)*log(2.)
    !Log(2 Gamma(n/2 +1) 2^(n/2))

   aset%calib_uncertainty = Ini_Read_Double('calib_uncertainty')
   aset%beam_uncertain = Ini_Read_logical('beam_uncertainty')
   window_dir  = Ini_Read_String('window_dir')
   Ini_fail_on_not_found = .false.
   windows_are_bare = Ini_Read_Logical('windows_are_bare',.false.)
   aset%file_points = Ini_read_Int('file_points',aset%num_points)

!Read in the observed values, errors and beam uncertainties
   allocate(aset%points(aset%num_points))
    Ini_fail_on_not_found = .true.
   do i=1, aset%num_points
      InLine = Ini_Read_String(numcat('data',i))
      if (aset%beam_uncertain) then
        read(InLine,*) aset%points(i)%obs, aset%points(i)%err_minus, aset%points(i)%err_plus, &
                        aset%points(i)%beam_err
        aset%points(i)%beam_err = aset%points(i)%beam_err/aset%points(i)%obs
      else
       read(InLine,*) aset%points(i)%obs, aset%points(i)%err_minus, aset%points(i)%err_plus
       aset%points(i)%beam_err = 0
      end if
      aset%points(i)%sigma = (aset%points(i)%err_minus + aset%points(i)%err_plus)/2
      aset%points(i)%var = aset%points(i)%sigma**2

      call ReadWindow(aset%points(i),trim(window_dir)//'/'//trim(numcat(aset%name,i)),windows_are_bare)
   end do

!See if the inverse covariance matrix is given (otherwise assume diagonal)
   Ini_fail_on_not_found = .false.


   Ninv_file = Ini_Read_String('N_inv')
   aset%has_corr_errors = Ninv_file /= ''
   if (aset%has_corr_errors) then
      allocate(tmp_mat(aset%file_points,aset%file_points))
      allocate(aset%N_inv(aset%num_points,aset%num_points))
      call ReadMatrix(Ninv_file, tmp_mat,aset%file_points,aset%file_points)
      aset%N_inv = tmp_mat(1:aset%num_points,1:aset%num_points)
      deallocate(tmp_mat)
   end if

   aset%use_hyperparameter = Ini_Read_Logical('use_hyperparameter',.false.)

!See if xfactors are given

   xfact_file = Ini_Read_String('xfactors')
   aset%has_xfactors = xfact_file /= ''

   if (aset%has_xfactors) then
      allocate(aset%xfactors(aset%num_points))
      call ReadVector(xfact_file, aset%xfactors,aset%num_points)
    
      aset%points(:)%var = aset%points(:)%var/(aset%points(:)%obs +aset%xfactors)**2
      aset%points(:)%obs = log(aset%points(:)%obs +aset%xfactors)
   end if
   call Ini_Close

   datasets(num_datasets) = aset

 end subroutine ReadDataset

 function GetWinBandPower(AP, cl)
   real  GetWinBandPower
   real cl(lmax)
   Type(CMBdatapoint) AP
   integer l
   real bandpower

   bandpower = 0
   do l= AP%win_min, AP%win_max
        bandpower = bandpower + (2*l+1)*cl(l)*AP%window(l)
   end do
   GetWinBandPower  = bandpower/(2*pi)
     
 end function GetWinBandPower
 
 subroutine InitNumericalMarge
   integer i

   do i= -halfsteps, halfsteps
    margeweights(i) = exp(-(i*3/real(halfsteps))**2/2)
   end do
   margenorm = sum(margeweights)

 end subroutine InitNumericalMarge

 function GetCalibMargexChisq(bandpowers, aset)
  !Numerically integrate over the calibration uncertainty
  !Assume Gaussian prior, as for analytic calculation without x-factors
  !Could also Monte-Carlo
   real GetCalibMargexChisq
   Type(CMBdataset) aset
   real bandpowers(aset%num_points),beambandpowers(aset%num_points),diffs(aset%num_points)
   real calib, chisq(-halfsteps:halfsteps),chisqcalib(-halfsteps:halfsteps)
   real minchisq
   integer i, ibeam

   if (margenorm == 0) call InitNumericalMarge

   do ibeam= -halfsteps, halfsteps

      if (aset%beam_uncertain) then
         beambandpowers = bandpowers*(1 + aset%points(:)%beam_err*ibeam*3/real(halfsteps))!Go out to 3 sigma
      else
         beambandpowers = bandpowers
      end if

      do i=-halfsteps,halfsteps

         calib = 1 + aset%calib_uncertainty*i*3./halfsteps !Go out to 3 sigma
         diffs(:) = aset%points(:)%obs- log(calib*beambandpowers(:) + aset%xfactors(:))
         if (aset%has_corr_errors) then
            chisq(i) = SUM(diffs*MATMUL(aset%N_inv,diffs))
         else
            chisq(i) = SUM(diffs**2/aset%points(:)%var)  
         end if
      end do

      minchisq = minval(chisq)

      chisqcalib(ibeam) = -2*log(sum(margeweights*exp(max(-30.,-(chisq-minchisq)/2)))/margenorm) + minchisq

      if (.not. aset%beam_uncertain) then
         GetCalibMargexChisq = chisqcalib(ibeam)
         return
      end if

   end do

   minchisq = minval(chisqcalib)
   GetCalibMargexChisq = -2*log(sum(margeweights*exp(max(-30.,-(chisqcalib-minchisq)/2)))/margenorm) + minchisq
  
 end function GetCalibMargexChisq

 function CalcLnLike(cl, aset) 
  !Compute -ln(Likelihood) 
   real cl(lmax), CalcLnLike
   Type(CMBdataset) aset
   integer i, j
   real chisq, bandpower
   real chi2op, chi2pp, wpp, wdd, Ninv
   real chi2dd,chi2pd,chi2od, b
   real bandpowers(aset%num_points), diffs(aset%num_points), tmp(aset%num_points), beam(aset%num_points)
   logical calib_marge
   real denom, calib

   if (.not. aset%use_set) then
      CalcLnLike = 0
      return
   end if
   
   denom = 1 !Assume Prob \propto exp(-chisq/2)/sqrt(denom)

   do i=1, aset%num_points
      bandpowers(i) = GetWinBandPower(aset%points(i), cl)
   end do
   
   if (aset%has_xfactors .and. (aset%calib_uncertainty > 1e-4 .or. aset%beam_uncertain))  then

      chisq = GetCalibMargexChisq(bandpowers,aset)

   else

    do i=1, aset%num_points
      if (aset%has_xfactors) then
       !obs in this case is Z = log(observed + x)
       diffs(i) = aset%points(i)%obs- log(bandpowers(i) + aset%xfactors(i))
      else
       diffs(i) = aset%points(i)%obs - bandpowers(i)
      end if
   end do

   if (aset%has_corr_errors) then
      chisq = SUM(diffs*MATMUL(aset%N_inv,diffs))
   else
      chisq = SUM(diffs**2/aset%points(:)%var)  
   end if

   if (aset%calib_uncertainty > 1e-4 .or. aset%beam_uncertain) then
 
     if (aset%has_corr_errors) then
        tmp = MATMUL(aset%N_inv,bandpowers)
     else
        tmp = bandpowers/aset%points(:)%var  
     end if
     chi2op = SUM(diffs*tmp)
     chi2pp = SUM(bandpowers*tmp)
     if (aset%beam_uncertain) then
         beam = aset%points(:)%beam_err*bandpowers
         if (aset%has_corr_errors) then
            tmp = MATMUL(aset%N_inv,beam)
         else
            tmp = beam/aset%points(:)%var 
         end if
         chi2dd = SUM(beam*tmp)
         chi2pd = SUM(bandpowers*tmp)
         chi2od = SUM(diffs*tmp)
     end if

     if (aset%calib_uncertainty > 1e-4) then
      !analytic marginalization over calibration uncertainty
      wpp = 1/(chi2pp+1/aset%calib_uncertainty**2) 
      chisq = chisq - wpp*chi2op**2 
      denom = denom/wpp*aset%calib_uncertainty**2
     else
      wpp = 0
     end if

     if (aset%beam_uncertain) then
      !analytic marginalization over beam uncertainty  
      wdd=1/(chi2dd-wpp*chi2pd**2+1)
      chisq = chisq - wdd*(chi2od-wpp*chi2op*chi2pd)**2
      denom = denom/wdd
     end if

   end if

   end if

  if (CMB_hyperparameters .or. aset%use_hyperparameter) then
    CalcLnLike = ((aset%num_points+2)*log(chisq+2) + log(denom))/2 - aset%hyper_norm
     !include the hyper_norm so as to compute relative evidences with and without hyperparameters
  else
    if (denom /= 1) chisq = chisq + log(denom)
    CalcLnLike = chisq/2
  end if
  
 end function CalcLnLike

 function CMBLnLike(cl)
  real cl(lmax), CMBLnLike
  integer i
  real tot(num_datasets)


  !$OMP PARALLEL DO DEFAULT(SHARED),SCHEDULE(STATIC), PRIVATE(i)
  do i=1, num_datasets
     tot(i) = CalcLnLike(cl,datasets(i))
  end do
  !$OMP END PARALLEL DO
  CMBLnLike = SUM(tot)

 end function

end module cmbdata


